/**
 * The Java serialization-compatible protocol implementation package.  Construct an instance of the {@link org.jboss.marshalling.serialization.java.JavaSerializationMarshallerFactory JavaSerializationMarshallerFactory}
 * class to use this marshalling protocol.
 */
package org.jboss.marshalling.serialization.java;