/**
 * The River protocol implementation package.  Construct an instance of the {@link org.jboss.marshalling.river.RiverMarshallerFactory RiverMarshallerFactory}
 * class to use the River marshalling protocol.
 */
package org.jboss.marshalling.river;
